#! /usr/bin/env python3
# encoding: utf-8

"""
Test manipulating and interogating trees using python modules with no external application dependancies.
"""

import unittest
import shutil
import os
import filecmp
from gubbins import common, treebuilders


class TestTreeMethods(unittest.TestCase):

    def test_robinson_foulds_distance(self):
        # two tree with different distances
        assert common.robinson_foulds_distance('gubbins/tests/data/robinson_foulds_distance_tree1.tre',
                                               'gubbins/tests/data/robinson_foulds_distance_tree2.tre') == 17.263494
        # two trees with same distance
        assert common.robinson_foulds_distance('gubbins/tests/data/robinson_foulds_distance_tree1.tre',
                                               'gubbins/tests/data/robinson_foulds_distance_tree1.tre') == 0

    def test_has_tree_been_seen_before(self):
        # same content so the tree has been seen before
        assert common.has_tree_been_seen_before(['gubbins/tests/data/robinson_foulds_distance_tree1.tre',
                                                 'gubbins/tests/data/robinson_foulds_distance_tree1.tre',
                                                 'gubbins/tests/data/robinson_foulds_distance_tree1_dup.tre'],
                                                'weighted_robinson_foulds') == 1
        # different trees
        assert common.has_tree_been_seen_before(['gubbins/tests/data/robinson_foulds_distance_tree1.tre',
                                                 'gubbins/tests/data/robinson_foulds_distance_tree1.tre',
                                                 'gubbins/tests/data/robinson_foulds_distance_tree2.tre'],
                                                'weighted_robinson_foulds') == 0

    def test_root_tree(self):
        common.root_tree('gubbins/tests/data/unrooted_tree.newick', 'gubbins/tests/data/actual_rooted_tree.newick')
        assert filecmp.cmp('gubbins/tests/data/actual_rooted_tree.newick',
                           'gubbins/tests/data/expected_rooted_tree.newick', shallow=False)
        os.remove('gubbins/tests/data/actual_rooted_tree.newick')

    def test_reroot_tree(self):
        shutil.copyfile('gubbins/tests/data/robinson_foulds_distance_tree1.tre',
                        'gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_actual')
        common.reroot_tree('gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_actual',
                           'sequence_4')
        assert filecmp.cmp('gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_actual',
                           'gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4')
        os.remove('gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_actual')

        shutil.copyfile('gubbins/tests/data/robinson_foulds_distance_tree1.tre',
                        'gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_actual')
        common.reroot_tree('gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_actual', '')
        assert filecmp.cmp('gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_actual',
                           'gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_expected')
        os.remove('gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_actual')

    def test_reroot_tree_with_outgroup(self):
        shutil.copyfile('gubbins/tests/data/robinson_foulds_distance_tree1.tre',
                        'gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_actual')
        common.reroot_tree_with_outgroup(
            'gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_actual', ['sequence_4'])
        assert filecmp.cmp('gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_actual',
                           'gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4')
        os.remove('gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_actual')

    def test_reroot_tree_with_outgroups(self):
        shutil.copyfile('gubbins/tests/data/robinson_foulds_distance_tree1.tre',
                        'gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_actual')
        common.reroot_tree_with_outgroup(
            'gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_actual',
            ['sequence_4', 'sequence_2'])
        assert filecmp.cmp('gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_actual',
                           'gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_2')
        os.remove('gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_actual')

    def test_reroot_tree_with_outgroups_all_in_one_clade(self):
        outgroups = ['A', 'B']
        expected_monophyletic_outgroup = ['A', 'B']
        expected_output_file = 'gubbins/tests/data/expected_reroot_tree_with_outgroups_all_in_one_clade.tre'
        self.reroot_tree_check(outgroups, expected_output_file, expected_monophyletic_outgroup)

    def test_reroot_tree_with_outgroups_all_in_one_clade_large(self):
        outgroups = ['A', 'B', 'C']
        expected_monophyletic_outgroup = ['A', 'B', 'C']
        expected_output_file = 'gubbins/tests/data/expected_test_reroot_tree_with_outgroups_all_in_one_clade_large.tre'
        self.reroot_tree_check(outgroups, expected_output_file, expected_monophyletic_outgroup)

    def test_reroot_tree_with_outgroups_all_in_different_clade(self):
        outgroups = ['A', 'D']
        expected_monophyletic_outgroup = ['A']
        expected_output_file = 'gubbins/tests/data/expected_reroot_tree_with_outgroups_all_in_different_clade.tre'
        self.reroot_tree_check(outgroups, expected_output_file, expected_monophyletic_outgroup)

    def test_reroot_tree_with_outgroups_with_two_mixed_clades(self):
        outgroups = ['A', 'B', 'C', 'D']
        expected_monophyletic_outgroup = ['A']
        expected_output_file = 'gubbins/tests/data/expected_reroot_tree_with_outgroups_with_two_mixed_clades.tre'
        self.reroot_tree_check(outgroups, expected_output_file, expected_monophyletic_outgroup)

    def reroot_tree_check(self, outgroups, expected_output_file, expected_monophyletic_outgroup):
        shutil.copyfile('gubbins/tests/data/outgroups_input.tre', '.tmp.outgroups_input.tre')
        assert expected_monophyletic_outgroup == common.get_monophyletic_outgroup('.tmp.outgroups_input.tre', outgroups)
        common.reroot_tree_with_outgroup('.tmp.outgroups_input.tre', outgroups)
        assert filecmp.cmp('.tmp.outgroups_input.tre', expected_output_file)
        os.remove('.tmp.outgroups_input.tre')

    def test_split_all_non_bi_nodes(self):
        # best way to access it is via reroot_tree_at_midpoint because it outputs to a file
        shutil.copyfile('gubbins/tests/data/non_bi_tree.tre', 'gubbins/tests/data/non_bi_tree.tre.actual')
        common.reroot_tree_at_midpoint('gubbins/tests/data/non_bi_tree.tre.actual')
        assert filecmp.cmp('gubbins/tests/data/non_bi_tree.tre.actual', 'gubbins/tests/data/non_bi_tree.tre.expected')
        os.remove('gubbins/tests/data/non_bi_tree.tre.actual')

    def test_reroot_tree_at_midpoint(self):
        shutil.copyfile('gubbins/tests/data/robinson_foulds_distance_tree1.tre',
                        'gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_actual')
        common.reroot_tree_at_midpoint(
            'gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_actual')
        assert filecmp.cmp('gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_actual',
                           'gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_expected')
        os.remove('gubbins/tests/data/robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_actual')

    def test_filter_out_removed_taxa_from_tree(self):
        common.filter_out_removed_taxa_from_tree('gubbins/tests/data/robinson_foulds_distance_tree1.tre',
             'gubbins/tests/data/robinson_foulds_distance_tree1.tre.filter_out_removed_taxa_from_tree_actual',
             ['sequence_1', 'sequence_2', 'sequence_3', 'sequence_4', 'sequence_5'])
        assert filecmp.cmp(
            'gubbins/tests/data/robinson_foulds_distance_tree1.tre.filter_out_removed_taxa_from_tree_actual',
            'gubbins/tests/data/robinson_foulds_distance_tree1.tre.filter_out_removed_taxa_from_tree_expected')
        os.remove('gubbins/tests/data/robinson_foulds_distance_tree1.tre.filter_out_removed_taxa_from_tree_actual')

    def test_internal_node_taxons_removed_when_used_as_starting_tree(self):
        common.filter_out_removed_taxa_from_tree('gubbins/tests/data/tree_with_internal_nodes.tre',
                                                 'gubbins/tests/data/tree_with_internal_nodes.tre_actual', [])
        assert filecmp.cmp('gubbins/tests/data/tree_with_internal_nodes.tre_actual',
                           'gubbins/tests/data/tree_with_internal_nodes.tre_expected')
        os.remove('gubbins/tests/data/tree_with_internal_nodes.tre_actual')

    def test_transfer_internal_node_labels_to_tree(self):
        reconstructor = treebuilders.IQTree(1, "internal_")
        common.transfer_internal_node_labels_to_tree('gubbins/tests/data/source_tree.tre',
                                                     'gubbins/tests/data/destination_tree.tre',
                                                     'gubbins/tests/data/renamed_output_tree', reconstructor)
        assert os.path.exists('gubbins/tests/data/renamed_output_tree')
        assert filecmp.cmp('gubbins/tests/data/renamed_output_tree',
                           'gubbins/tests/data/expected_renamed_output_tree', shallow=False)
        os.remove('gubbins/tests/data/renamed_output_tree')

    def test_remove_internal_node_labels(self):
        common.remove_internal_node_labels_from_tree('gubbins/tests/data/final_tree_with_internal_labels.tre',
                                                     'final_tree_with_internal_labels.tre')
        assert os.path.exists('final_tree_with_internal_labels.tre')
        assert filecmp.cmp('final_tree_with_internal_labels.tre',
                           'gubbins/tests/data/expected_final_tree_without_internal_labels.tre')
        os.remove('final_tree_with_internal_labels.tre')


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

message = ("\nNOTE: gubbins_drawer.py is no longer supported. "
           "Instead we recommend using Phandango[1] (http://phandango.net/) to visualise the outputs of Gubbins.\n\n"
           "[1] James Hadfield, Nicholas J Croucher, Richard J Goater, Khalil Abudahab, David M Aanensen, Simon R Harris. "
           "\"Phandango: an interactive viewer for bacterial population genomics\". "
           "Bioinformatics, Volume 34, Issue 2, 15 January 2018, Pages 292–293, "
           "https://doi.org/10.1093/bioinformatics/btx610\n")
print(message)



#! /usr/bin/env python3

"""
Imports into the `gubbins` namespace all fundamental
classes and methods for instantiating objects in the
for usage by client code.
"""

import sys
import os

###############################################################################
## Populate the 'gubbins' namespace

from gubbins import common

###############################################################################
## PACKAGE METADATA

__project__ = "Gubbins"
__version__ = "0.1"

try:
    try:
        __homedir__ = __path__[0]
    except AttributeError:
        __homedir__ = os.path.dirname(os.path.abspath(__file__))
    except IndexError:
        __homedir__ = os.path.dirname(os.path.abspath(__file__))
except OSError:
    __homedir__ = None
except:
    __homedir__ = None

__author__ = "Andrew J. Page, Nicholas Croucher, Aidan Delaney and Simon Harris"
__copyright__ = "Copyright 2013 Wellcome Trust Sanger Institutue"
__license__ = """
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
PACKAGE_VERSION = __version__

def description():
    if __revision__.is_available:
        revision_text = " (%s)" % str(__revision__)
    else:
        revision_text = ""
    return "%s %s%s" % (__project__, __version__, revision_text)

if __name__ == "__main__":
    sys.stdout.write("%s\n" % description())




#!/usr/bin/env python3
# encoding: utf-8
#
# Wellcome Trust Sanger Institute
# Copyright (C) 2012  Wellcome Trust Sanger Institute
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import argparse
import pkg_resources
import gubbins.common


def main():

    parser = argparse.ArgumentParser(
        description='Croucher N. J., Page A. J., Connor T. R., Delaney A. J., Keane J. A., Bentley S. D., Parkhill J., '
                    'Harris S.R. "Rapid phylogenetic analysis of large samples of recombinant bacterial whole genome '
                    'sequences using Gubbins". Nucleic Acids Res. 2015 Feb 18;43(3):e15. doi: 10.1093/nar/gku1196.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('alignment_filename',        help='Multifasta alignment file')
    parser.add_argument('--outgroup',          '-o', help='Outgroup name for rerooting. A list of comma separated '
                                                          'names can be used if they form a clade')
    parser.add_argument('--starting_tree',     '-s', help='Starting tree')
    parser.add_argument('--use_time_stamp',    '-u', help='Use a time stamp in file names', action='store_true')
    parser.add_argument('--verbose',           '-v', help='Turn on debugging', action='store_true')
    parser.add_argument('--no_cleanup',        '-n', help='Dont cleanup intermediate files', action='store_true')
    parser.add_argument('--tree_builder',      '-t', help='Application to use for tree building', default="raxml",
                        choices=['raxml', 'fasttree', 'hybrid'])
    parser.add_argument('--iterations',        '-i', help='Maximum No. of iterations', type=int, default=5)
    parser.add_argument('--min_snps',          '-m', help='Min SNPs to identify a recombination block', type=int,
                        default=3)
    parser.add_argument('--filter_percentage', '-f', help='Filter out taxa with more than this percentage of gaps',
                        type=int, default=25)
    parser.add_argument('--prefix',            '-p', help='Add a prefix to the final output filenames')
    parser.add_argument('--threads',           '-c', help='Number of threads to run with RAXML, but only if a PTHREADS '
                                                         'version is available', type=int,  default=1)
    parser.add_argument('--converge_method',   '-z', help='Criteria to use to know when to halt iterations',
                        default='weighted_robinson_foulds', choices=['weighted_robinson_foulds', 'robinson_foulds',
                                                                     'recombination'])
    parser.add_argument('--version',                 action='version',
                        version=str(pkg_resources.get_distribution("gubbins").version))
    parser.add_argument('--min_window_size',   '-a', help='Minimum window size', type=int, default=100)
    parser.add_argument('--max_window_size',   '-b', help='Maximum window size', type=int, default=10000)
    parser.add_argument('--raxml_model',       '-r', help='RAxML model', default='GTRCAT',
                        choices=['GTRGAMMA', 'GTRCAT'])
    parser.add_argument('--remove_identical_sequences', '-d', help='Remove identical sequences', action='store_true')

    gubbins.common.parse_and_run(parser.parse_args(), parser.description)



#! /usr/bin/env python3
# encoding: utf-8

"""
Tests of utilities.
"""

import unittest
import os
import subprocess
import re
import io
from contextlib import redirect_stdout
from gubbins import common, utils

modules_dir = os.path.dirname(os.path.abspath(utils.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestUtilities(unittest.TestCase):

    def test_gubbins_command(self):
        assert common.create_gubbins_command('AAA', 'BBB', 'CCC', 'DDD', 'EEE', 5, 10, 200) \
               == 'AAA -r -v CCC -a 10 -b 200 -f EEE -t DDD -m 5 BBB'

    def test_translation_of_filenames_to_final_filenames(self):
        assert common.translation_of_filenames_to_final_filenames('AAA', 'test') == {
            'AAA.vcf':             'test.summary_of_snp_distribution.vcf',
            'AAA.branch_snps.tab': 'test.branch_base_reconstruction.embl',
            'AAA.tab':             'test.recombination_predictions.embl',
            'AAA.gff':             'test.recombination_predictions.gff',
            'AAA.stats':           'test.per_branch_statistics.csv',
            'AAA.snp_sites.aln':   'test.filtered_polymorphic_sites.fasta',
            'AAA.phylip':          'test.filtered_polymorphic_sites.phylip',
            'AAA.internal':        'test.node_labelled.final_tree.tre',
            'AAA':                 'test.final_tree.tre'
        }

    def test_check_and_fix_window_size(self):
        class ArgsObject:
            pass
        arguments = ArgsObject
        arguments.min_window_size = 1
        arguments.max_window_size = 2000000
        common.check_and_fix_window_size(arguments)
        assert arguments.min_window_size == 3
        assert arguments.max_window_size == 1000000
        arguments.min_window_size = 20
        arguments.max_window_size = 10
        common.check_and_fix_window_size(arguments)
        assert arguments.min_window_size == 10
        assert arguments.max_window_size == 20

    def test_which(self):
        # the location of ls varies depending on OS so just check end
        assert re.match('.*/ls$', utils.which('ls')) is not None
        # Strip parameters
        assert re.match('.*/ls$', utils.which('ls -alrt')) is not None
        assert utils.which('non_existent_program') is None

    def test_is_executable(self):
        program = utils.which('ls')
        assert utils.is_executable(program)
        assert not utils.is_executable('non_existent_program')

    def test_replace_executable(self):
        assert 'raxmlHPC -f d -p 1 -m GTRGAMMA' in utils.replace_executable('raxml -f d -p 1 -m GTRGAMMA', 'raxmlHPC')
        assert '../src/gubbins' in utils.replace_executable('gubbins', '../src/gubbins')

    def test_rename_files(self):
        subprocess.check_call('touch temp_file; touch another_file', shell=True)
        utils.rename_files({'temp_file': 'output_file', 'another_file': 'another_output_file'})
        assert os.path.exists('output_file')
        assert os.path.exists('another_output_file')
        os.remove('output_file')
        os.remove('another_output_file')

    def test_delete_files(self):
        open(os.path.join(data_dir, 'AAA.rex'), 'w').close()
        open(os.path.join(data_dir, 'BBB.rex'), 'w').close()
        open(os.path.join(data_dir, 'AAA.rox'), 'w').close()
        open(os.path.join(data_dir, 'BBB.rox'), 'w').close()
        utils.delete_files(data_dir, ['AAA'], ".r[eo]x")
        assert not os.path.exists(os.path.join(data_dir, 'AAA.rex'))
        assert not os.path.exists(os.path.join(data_dir, 'AAA.rox'))
        assert os.path.exists(os.path.join(data_dir, 'BBB.rex'))
        assert os.path.exists(os.path.join(data_dir, 'BBB.rox'))
        os.remove(os.path.join(data_dir, 'BBB.rex'))
        os.remove(os.path.join(data_dir, 'BBB.rox'))

    def test_do_files_exist(self):
        open(os.path.join(data_dir, 'AAA.rex'), 'w').close()
        open(os.path.join(data_dir, 'BBB.rox'), 'w').close()
        assert utils.do_files_exist(data_dir, ['AAA', 'BBB'], ".r[eo]x")
        assert not utils.do_files_exist(data_dir, ['AAA', 'BBB'], ".rax")
        os.remove(os.path.join(data_dir, 'AAA.rex'))
        os.remove(os.path.join(data_dir, 'BBB.rox'))

    def test_verbose_printer(self):
        printer = utils.VerbosePrinter(True, "-")
        assert printer.is_verbose()
        assert printer.separator() == "-"
        f = io.StringIO()
        with redirect_stdout(f):
            printer.print(["AAA", "BBB"])
            printed = f.getvalue()
        assert printed == "AAA-BBB\n"



# encoding: utf-8

"""
Tests for preprocessing the input FASTA file
"""

import unittest
import re
import os
import subprocess
import filecmp
import pprint
from gubbins.PreProcessFasta import PreProcessFasta

class TestPreProcessFasta(unittest.TestCase):
      
  def test_input_file_with_no_duplicate_sequences(self):   
      preprocessfasta = PreProcessFasta('gubbins/tests/data/preprocessfasta/no_duplicates.aln')
      self.assertEqual(preprocessfasta.hash_sequences(), 
       {b"\x840\x89L\xfe\xb5J6%\xf1\x8f\xe2O\xce'.": ['sample1'],
        b'\x9c\xe6\x8b\xf7\xae\xe2\x1f\xf5j\xcfu\xf4\xfdO\x8b\xec': ['sample4'],
        b'\xc3n`\xf5t\x00\x1e\xf3\xde\nU\x1f\x95\x0b\xdb9': ['sample2'],
        b'\xdf\xedM\xdf1\xf2PPO\xc1\xf54T?\xdb\xa2': ['sample3']})
      self.assertEqual(preprocessfasta.calculate_sequences_missing_data_percentage(), {'sample1': 0.0,
 'sample2': 0.0,
 'sample3': 0.0,
 'sample4': 0.0})
        
      self.assertEqual(preprocessfasta.taxa_of_duplicate_sequences(),[])

      preprocessfasta.remove_duplicate_sequences_and_sequences_missing_too_much_data('output.aln',1)
      self.assertTrue(filecmp.cmp('output.aln', 'gubbins/tests/data/preprocessfasta/no_duplicates.aln'))

  def test_input_file_with_one_duplicate_sequences(self):   
      preprocessfasta = PreProcessFasta('gubbins/tests/data/preprocessfasta/one_duplicate.aln')
      self.assertEqual(preprocessfasta.hash_sequences(), 
       {b"\x840\x89L\xfe\xb5J6%\xf1\x8f\xe2O\xce'.": ['sample1', 'sample3'],
        b'\x9c\xe6\x8b\xf7\xae\xe2\x1f\xf5j\xcfu\xf4\xfdO\x8b\xec': ['sample4'],
        b'\xc3n`\xf5t\x00\x1e\xf3\xde\nU\x1f\x95\x0b\xdb9': ['sample2']})
        
      self.assertEqual(preprocessfasta.taxa_of_duplicate_sequences(),['sample1'])
      
      preprocessfasta.remove_duplicate_sequences_and_sequences_missing_too_much_data('output.aln',1)
      self.assertTrue(filecmp.cmp('output.aln', 'gubbins/tests/data/preprocessfasta/expected_one_duplicate.aln'))
 
  def test_input_file_with_multiple_duplicate_sequences(self):   
      preprocessfasta = PreProcessFasta('gubbins/tests/data/preprocessfasta/multiple_duplicates.aln')
      self.assertEqual(preprocessfasta.hash_sequences(), 
       {b"\x840\x89L\xfe\xb5J6%\xf1\x8f\xe2O\xce'.": ['sample1', 'sample3'],
        b'\x9c\xe6\x8b\xf7\xae\xe2\x1f\xf5j\xcfu\xf4\xfdO\x8b\xec': ['sample2', 'sample4']})
        
      self.assertEqual(preprocessfasta.taxa_of_duplicate_sequences(),['sample1','sample2'])
      
      preprocessfasta.remove_duplicate_sequences_and_sequences_missing_too_much_data('output.aln',1)
      self.assertTrue(filecmp.cmp('output.aln', 'gubbins/tests/data/preprocessfasta/expected_multiple_duplicates.aln'))
      
      
  def test_dont_filter_input_file_with_multiple_duplicate_sequences(self):   
      preprocessfasta = PreProcessFasta('gubbins/tests/data/preprocessfasta/multiple_duplicates.aln')
      self.assertEqual(preprocessfasta.hash_sequences(), 
       {b"\x840\x89L\xfe\xb5J6%\xf1\x8f\xe2O\xce'.": ['sample1', 'sample3'],
        b'\x9c\xe6\x8b\xf7\xae\xe2\x1f\xf5j\xcfu\xf4\xfdO\x8b\xec': ['sample2', 'sample4']})
        
      self.assertEqual(preprocessfasta.taxa_of_duplicate_sequences(),['sample1','sample2'])
      
      preprocessfasta.remove_duplicate_sequences_and_sequences_missing_too_much_data('output.aln',0)
      self.assertTrue(filecmp.cmp('output.aln', 'gubbins/tests/data/preprocessfasta/multiple_duplicates.aln'))    
      
      
 
  def test_input_file_with_all_duplicate_sequences(self):   
      preprocessfasta = PreProcessFasta('gubbins/tests/data/preprocessfasta/all_same_sequence.aln')
      self.assertEqual(preprocessfasta.hash_sequences(), 
       {b"\x840\x89L\xfe\xb5J6%\xf1\x8f\xe2O\xce'.": ['sample1',
                                                      'sample2',
                                                      'sample3',
                                                      'sample4']})
                                                      
      self.assertEqual(preprocessfasta.taxa_of_duplicate_sequences(),['sample1',
                                                      'sample2',
                                                      'sample3'])
                                                      
  def test_filter_out_alignments_with_too_much_missing_data(self):
    preprocessfasta = PreProcessFasta('gubbins/tests/data/preprocessfasta/missing_data.aln', False, 5)
    preprocessfasta.remove_duplicate_sequences_and_sequences_missing_too_much_data('output.aln',1)
    self.assertTrue(filecmp.cmp('output.aln','gubbins/tests/data/preprocessfasta/expected_missing_data.aln'))        
      
  def tearDown(self):
      for file_to_delete in ['output.aln']:
          if os.path.exists(file_to_delete):
              os.remove(file_to_delete) 

if __name__ == "__main__":
  unittest.main(buffer=True)



#! /usr/bin/env python3
# encoding: utf-8

"""
Integration testing of external dependencies. Likely to be the most brittle tests, but the most important.
"""

import unittest
import os
import glob
import argparse
import pkg_resources
from gubbins import common

modules_dir = os.path.dirname(os.path.abspath(common.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestExternalDependencies(unittest.TestCase):

    def test_fasttree(self):
        self.assertEqual(2, 1+1)
        parser = self.default_arg_parse()
        common.parse_and_run(parser.parse_args(["--tree_builder", "fasttree",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')

    def test_iqtree(self):
        parser = self.default_arg_parse()
        common.parse_and_run(parser.parse_args(["--tree_builder", "iqtree",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')

    def test_raxml(self):
        parser = self.default_arg_parse()
        common.parse_and_run(parser.parse_args(["--tree_builder", "raxml",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')

    def test_hybrid(self):
        parser = self.default_arg_parse()
        common.parse_and_run(parser.parse_args(["--tree_builder", "hybrid",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')

    def test_rename_final_output(self):
        parser = self.default_arg_parse()
        common.parse_and_run(parser.parse_args(["--prefix", "different_prefix",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        self.check_for_output_files('different_prefix')
        self.cleanup('different_prefix')

    def test_cleanup(self):
        parser = self.default_arg_parse()
        common.parse_and_run(parser.parse_args([os.path.join(data_dir, 'multiple_recombinations.aln')]))
        assert not glob.glob('multiple_recombinations.aln.*')
        assert not glob.glob('multiple_recombinations.iteration*')
        self.cleanup('multiple_recombinations')

    @staticmethod
    def check_for_output_files(prefix):
        assert os.path.exists(prefix + '.summary_of_snp_distribution.vcf')
        assert os.path.exists(prefix + '.recombination_predictions.embl')
        assert os.path.exists(prefix + '.per_branch_statistics.csv')
        assert os.path.exists(prefix + '.filtered_polymorphic_sites.fasta')
        assert os.path.exists(prefix + '.filtered_polymorphic_sites.phylip')
        assert os.path.exists(prefix + '.recombination_predictions.gff')
        assert os.path.exists(prefix + '.branch_base_reconstruction.embl')
        assert os.path.exists(prefix + '.final_tree.tre')
        assert os.path.exists(prefix + '.node_labelled.final_tree.tre')

    @staticmethod
    def cleanup(prefix):
        regex_to_remove = prefix + ".*"
        for file in glob.glob(regex_to_remove):
            os.remove(file)

    @staticmethod
    def default_arg_parse():
        # Keep this up to date with run_gubbins.py!
        parser = argparse.ArgumentParser()
        parser.add_argument('alignment_filename', help='Multifasta alignment file')
        parser.add_argument('--outgroup', '-o', help='Outgroup name for rerooting. A list of comma separated names '
                                                     'can be used if they form a clade')
        parser.add_argument('--starting_tree', '-s', help='Starting tree')
        parser.add_argument('--use_time_stamp', '-u', help='Use a time stamp in file names', action='store_true')
        parser.add_argument('--verbose', '-v', help='Turn on debugging', action='store_true')
        parser.add_argument('--no_cleanup', '-n', help='Dont cleanup intermediate files', action='store_true')
        parser.add_argument('--tree_builder', '-t', help='Application to use for tree building', default="raxml",
                            choices=['raxml', 'fasttree', 'hybrid', 'iqtree'])
        parser.add_argument('--iterations', '-i', help='Maximum No. of iterations', type=int, default=5)
        parser.add_argument('--min_snps', '-m', help='Min SNPs to identify a recombination block', type=int, default=3)
        parser.add_argument('--filter_percentage', '-f', help='Filter out taxa with more than this percentage of gaps',
                            type=int, default=25)
        parser.add_argument('--prefix', '-p', help='Add a prefix to the final output filenames')
        parser.add_argument('--threads', '-c', help='Number of threads to run with RAXML, but only if a PTHREADS '
                                                    'version is available', type=int, default=1)
        parser.add_argument('--converge_method', '-z', help='Criteria to use to know when to halt iterations',
                            default='weighted_robinson_foulds', choices=['weighted_robinson_foulds', 'robinson_foulds',
                                                                         'recombination'])
        parser.add_argument('--version', action='version',
                            version=str(pkg_resources.get_distribution("gubbins").version))
        parser.add_argument('--min_window_size', '-a', help='Minimum window size', type=int, default=100)
        parser.add_argument('--max_window_size', '-b', help='Maximum window size', type=int, default=10000)
        parser.add_argument('--raxml_model', '-r', help='RAxML model', default='GTRCAT', choices=['GTRGAMMA', 'GTRCAT'])
        parser.add_argument('--remove_identical_sequences', '-d', help='Remove identical sequences',
                            action='store_true')
        return parser


if __name__ == "__main__":
    unittest.main(buffer=True)



#! /usr/bin/env python3
# encoding: utf-8

"""
Tests if the same recombinations are seen in multiple iterations
"""

import unittest
import os
from gubbins import common

modules_dir = os.path.dirname(os.path.abspath(common.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestConvergingRecombinations(unittest.TestCase):

    def test_reading_embl_file(self):
        assert common.extract_recombinations_from_embl(os.path.join(data_dir, 'small_recombination.embl')) \
               == {'sequence1': [[1, 3], [5, 7], [9, 11]], 'sequence2': [[1, 3], [9, 11]], 'sequence3': [[1, 3]]}

    def test_two_files_are_different(self):
        assert not common.have_recombinations_been_seen_before(os.path.join(data_dir, 'small_recombination.embl'),
                             [os.path.join(data_dir, 'small_recombination_different.embl')])

    def test_two_files_are_same(self):
        assert common.have_recombinations_been_seen_before(os.path.join(data_dir, 'small_recombination.embl'),
                                                           [os.path.join(data_dir, 'small_recombination.embl')])

    def test_multiple_files_have_one_match(self):
        assert common.have_recombinations_been_seen_before(os.path.join(data_dir, 'small_recombination.embl'),
            [os.path.join(data_dir, 'small_recombination_different.embl'),
             os.path.join(data_dir, 'small_recombination.embl'),
             os.path.join(data_dir, 'small_recombination_different.embl')])

    def test_get_recombination_files(self):
        assert common.get_recombination_files(['AAA', 'BBB', 'CCC']) == ('CCC.tab', ['AAA.tab', 'BBB.tab'])


if __name__ == "__main__":
    unittest.main()



# encoding: utf-8
# Wellcome Trust Sanger Institute
# Copyright (C) 2013  Wellcome Trust Sanger Institute
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import sys
from gubbins import utils


class FastTree:
    """Class for operations with the FastTree executable"""

    def __init__(self, verbose=False):
        """Initialises the object"""
        self.verbose = verbose
        self.tree_prefix = ""
        self.tree_suffix = ".tre"

        self.potential_executables = ["FastTree", "fasttree"]
        self.executable = utils.choose_executable(self.potential_executables)
        if self.executable is None:
            sys.exit("No usable version of FastTree could be found.")
        self.tree_building_parameters = ["-nosupport", "-gtr", "-gamma", "-nt"]

    def tree_building_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Constructs the command to call the FastTree executable"""
        output_tree = basename + self.tree_suffix
        command = [self.executable]
        command.extend(self.tree_building_parameters)
        if input_tree:
            command.extend(["-intree", input_tree])
        command.extend(["-out", output_tree])
        command.append(alignment_filename)
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)


class IQTree:
    """Class for operations with the IQTree executable"""

    def __init__(self, threads: int, internal_node_prefix="", verbose=False):
        """Initialises the object"""
        self.verbose = verbose
        self.threads = threads
        self.tree_prefix = ""
        self.tree_suffix = ".treefile"
        self.asr_prefix = ""
        self.asr_suffix = ".state"
        self.asr_tree_prefix = ""
        self.asr_tree_suffix = ".treefile"
        self.internal_node_prefix = internal_node_prefix

        self.executable = "iqtree"
        if utils.which(self.executable) is None:
            sys.exit("No usable version of IQTree could be found.")
        self.tree_building_parameters = ["-safe -m GTR+G4"]
        self.internal_sequence_reconstruction_parameters = ["-safe -asr -m GTR+G4"]

    def tree_building_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Constructs the command to call the IQTree executable"""
        command = [self.executable]
        command.extend(self.tree_building_parameters)
        command.extend(["-s", alignment_filename, "-pre", basename])
        if self.threads:
            command.extend(["-nt", str(self.threads)])
        else:
            command.extend(["-nt", "AUTO"])
        if input_tree:
            command.extend(["-t", input_tree])
        if not self.verbose:
            command.append("-quiet")
        return " ".join(command)

    def internal_sequence_reconstruction_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Constructs the command to call the IQTree executable for ancestral sequence reconstruction"""
        command = [self.executable]
        command.extend(self.internal_sequence_reconstruction_parameters)
        command.extend(["-s", alignment_filename, "-pre", basename])
        if self.threads:
            command.extend(["-nt", str(self.threads)])
        if input_tree:
            command.extend(["-te", input_tree])
        if not self.verbose:
            command.append("-quiet")
        return " ".join(command)

    def convert_raw_ancestral_states_to_fasta(self, input_filename, output_filename):
        """Converts the file containing ancestral sequences into FASTA format"""
        raw_sequences = {}
        with open(input_filename, 'r') as infile:
            for line in infile:
                if not line.startswith("#") and not line.startswith("Node\tSite\tState"):
                    elements = line.split("\t")
                    name = elements[0]
                    base = elements[2]
                    if name in raw_sequences:
                        raw_sequences[name].append(base)
                    else:
                        raw_sequences[name] = [base]

        with open(output_filename, 'w+') as outfile:
            for sequence_name, sequence_bases in raw_sequences.items():
                outfile.write(">" + self.replace_internal_node_label(sequence_name) + "\n")
                outfile.write("".join(sequence_bases) + "\n")

    def replace_internal_node_label(self, label):
        """Changes the label of internal nodes"""
        return self.internal_node_prefix + label.replace("Node", "")


class RAxML:
    """Class for operations with the RAxML executable"""

    def __init__(self, threads: int, model='GTRCAT', internal_node_prefix="", verbose=False):
        """Initialises the object"""
        self.verbose = verbose
        self.threads = threads
        self.model = model
        self.tree_prefix = "RAxML_result."
        self.tree_suffix = ""
        self.asr_prefix = "RAxML_marginalAncestralStates."
        self.asr_suffix = ""
        self.asr_tree_prefix = "RAxML_nodeLabelledRootedTree."
        self.asr_tree_suffix = ""
        self.internal_node_prefix = internal_node_prefix

        self.single_threaded_executables = ['raxmlHPC-AVX2', 'raxmlHPC-AVX', 'raxmlHPC-SSE3', 'raxmlHPC']
        self.multi_threaded_executables = ['raxmlHPC-PTHREADS-AVX2', 'raxmlHPC-PTHREADS-AVX',
                                           'raxmlHPC-PTHREADS-SSE3', 'raxmlHPC-PTHREADS']
        self.executable = self.select_executable_based_on_threads()
        if self.executable is None:
            sys.exit("No usable version of RAxML could be found.")

        self.tree_building_parameters = ["-f", "d", "-p", str(1)]
        if self.model == "GTRGAMMA":
            self.tree_building_parameters.extend(["-m", "GTRGAMMA"])
        else:
            self.tree_building_parameters.extend(["-m", "GTRCAT", "-V"])
        self.internal_sequence_reconstruction_parameters = ["-f", "A", "-p", str(1), "-m", "GTRGAMMA"]

    def tree_building_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Constructs the command to call the RAxML executable for tree building"""
        command = [self.executable]
        command.extend(self.tree_building_parameters)
        command.extend(["-s", alignment_filename, "-n", basename])
        if self.threads > 1:
            command.extend(["-T", str(self.threads)])
        if input_tree:
            command.extend(["-t", input_tree])
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)

    def internal_sequence_reconstruction_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Constructs the command to call the RAxML executable for ancestral sequence reconstruction"""
        command = [self.executable]
        command.extend(self.internal_sequence_reconstruction_parameters)
        command.extend(["-s", alignment_filename, "-n", basename])
        if self.threads > 1:
            command.extend(["-T", str(self.threads)])
        command.extend(["-t", input_tree])
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)

    def select_executable_based_on_threads(self):
        """Chooses an appropriate executable"""
        if self.threads == 1:
            single_threaded_exec = utils.choose_executable_based_on_processor(
                self.single_threaded_executables)
            if single_threaded_exec is not None:
                return single_threaded_exec
            else:
                print("Trying multithreaded version of RAxML because no single threaded version of RAxML could be "
                      "found. Just to warn you, this requires 2 threads.\n")
                self.threads = 2

        if self.threads > 1:
            multi_threaded_exec = utils.choose_executable_based_on_processor(
                self.multi_threaded_executables)
            if multi_threaded_exec is not None:
                return multi_threaded_exec
            else:
                return None

    def convert_raw_ancestral_states_to_fasta(self, input_filename, output_filename):
        """Converts the file containing ancestral sequences into FASTA format"""
        with open(input_filename, 'r') as infile:
            with open(output_filename, 'w+') as outfile:
                for sequence_line in infile:
                    [sequence_name, sequence_bases] = sequence_line.split(' ')
                    sequence_bases = sequence_bases.replace('?', 'N')
                    outfile.write('>' + self.replace_internal_node_label(sequence_name) + '\n')
                    outfile.write(sequence_bases)

    def replace_internal_node_label(self, label):
        """Changes the label of internal nodes"""
        return self.internal_node_prefix + label



#! /usr/bin/env python3
# encoding: utf-8

"""
Tests alignment manipulation and conversion with no external application dependencies.
"""

import unittest
import filecmp
import os
from gubbins import common

modules_dir = os.path.dirname(os.path.abspath(common.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestAlignmentMethods(unittest.TestCase):

    def test_number_of_sequences_in_alignment(self):
        assert common.number_of_sequences_in_alignment(os.path.join(data_dir, 'small_alignment.aln')) == 5

    def test_get_sequence_names_from_alignment(self):
        assert common.get_sequence_names_from_alignment(os.path.join(data_dir, 'small_alignment.aln')) == \
               ['sequence1', 'sequence2', 'sequence3', 'sequence4', 'sequence5']

    def test_reinsert_gaps_into_fasta_file(self):
        common.reinsert_gaps_into_fasta_file(os.path.join(data_dir, 'gaps_to_be_reinserted.aln'),
                                             os.path.join(data_dir, 'gaps_to_be_reinserted.vcf'),
                                             os.path.join(data_dir, 'gaps_to_be_reinserted.aln.actual'))
        assert filecmp.cmp(os.path.join(data_dir, 'gaps_to_be_reinserted.aln.actual'),
                           os.path.join(data_dir, 'gaps_to_be_reinserted.aln.expected'))
        os.remove(os.path.join(data_dir, 'gaps_to_be_reinserted.aln.actual'))

    def test_reconvert_fasta_file(self):
        common.reconvert_fasta_file(os.path.join(data_dir, 'alignment_with_too_much_missing_data.aln'),
                                    os.path.join(data_dir, 'reconvert_fasta_file.aln.actual'))
        assert filecmp.cmp(os.path.join(data_dir, 'reconvert_fasta_file.aln.actual'),
                           os.path.join(data_dir, 'reconvert_fasta_file.aln.expected'))
        os.remove(os.path.join(data_dir, 'reconvert_fasta_file.aln.actual'))

    def test_concatenate_fasta_files(self):
        common.concatenate_fasta_files([os.path.join(data_dir, 'small_alignment.aln'),
                                        os.path.join(data_dir, 'further_alignment.aln')],
                                       os.path.join(data_dir, 'concatenate_fasta_files.aln.actual'))
        assert filecmp.cmp(os.path.join(data_dir, 'concatenate_fasta_files.aln.actual'),
                           os.path.join(data_dir, 'concatenate_fasta_files.aln.expected'))
        os.remove(os.path.join(data_dir, 'concatenate_fasta_files.aln.actual'))


if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python3
# encoding: utf-8

"""
Tests the construction of file names and commands with no dependencies on external applications.
"""

import unittest.mock
import os
import filecmp
from gubbins import treebuilders, utils

modules_dir = os.path.dirname(os.path.abspath(treebuilders.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestStringConstruction(unittest.TestCase):

    def test_fasttree_treebuilding_command(self):
        fasttree = treebuilders.FastTree()
        fasttree.executable = 'DDD'
        fasttree.tree_building_parameters = ['EEE']
        assert fasttree.tree_building_command('AAA', '', 'CCC') == 'DDD EEE -out CCC.tre AAA > /dev/null 2>&1'
        fasttree.verbose = True
        assert fasttree.tree_building_command('AAA', 'BBB', 'CCC') == 'DDD EEE -intree BBB -out CCC.tre AAA'

    def test_iqtree_treebuilding_command(self):
        iqtree = treebuilders.IQTree(8)
        iqtree.executable = 'DDD'
        iqtree.tree_building_parameters = ['EEE']
        assert iqtree.tree_building_command('AAA', '', 'CCC') == 'DDD EEE -s AAA -pre CCC -nt 8 -quiet'
        iqtree.verbose = True
        iqtree.threads = 0
        assert iqtree.tree_building_command('AAA', 'BBB', 'CCC') == 'DDD EEE -s AAA -pre CCC -nt AUTO -t BBB'

    def test_iqtree_sequence_reconstruction_command(self):
        iqtree = treebuilders.IQTree(8)
        iqtree.executable = 'DDD'
        iqtree.internal_sequence_reconstruction_parameters = ['EEE']
        assert iqtree.internal_sequence_reconstruction_command('AAA', 'BBB', 'CCC') \
               == 'DDD EEE -s AAA -pre CCC -nt 8 -te BBB -quiet'

    def test_iqtree_convert_raw_ancestral_states_to_fasta(self):
        iqtree = treebuilders.IQTree(8)
        iqtree.convert_raw_ancestral_states_to_fasta(os.path.join(data_dir, 'iqtree_ancestral.state'),
                                                     os.path.join(data_dir, 'iqtree_ancestral.fasta.actual'))
        assert filecmp.cmp(os.path.join(data_dir, 'iqtree_ancestral.fasta.actual'),
                           os.path.join(data_dir, 'iqtree_ancestral.fasta.expected'))
        os.remove(os.path.join(data_dir, 'iqtree_ancestral.fasta.actual'))

    def test_iqtree_replace_internal_node_label(self):
        iqtree = treebuilders.IQTree(8, 'AAA')
        assert iqtree.replace_internal_node_label('Node20') == 'AAA20'

    def test_raxml_treebuilding_command(self):
        raxml = treebuilders.RAxML(8)
        raxml.executable = 'DDD'
        raxml.tree_building_parameters = ['EEE']
        assert raxml.tree_building_command('AAA', '', 'CCC') == 'DDD EEE -s AAA -n CCC -T 8 > /dev/null 2>&1'
        raxml.verbose = True
        raxml.threads = 1
        assert raxml.tree_building_command('AAA', 'BBB', 'CCC') == 'DDD EEE -s AAA -n CCC -t BBB'

    def test_raxml_sequence_reconstruction_command(self):
        raxml = treebuilders.RAxML(8)
        raxml.executable = 'DDD'
        raxml.internal_sequence_reconstruction_parameters = ['EEE']
        assert raxml.internal_sequence_reconstruction_command('AAA', 'BBB', 'CCC') \
               == 'DDD EEE -s AAA -n CCC -T 8 -t BBB > /dev/null 2>&1'

    @unittest.mock.patch('gubbins.utils.choose_executable_based_on_processor')
    def test_raxml_select_executable_based_on_threads(self, mock_choice):
        self.assertIs(mock_choice, utils.choose_executable_based_on_processor)

        raxml = treebuilders.RAxML(8)
        raxml.single_threaded_executables = ['BBB', 'CCC']
        raxml.multi_threaded_executables = ['DDD', 'EEE']

        mock_choice.return_value = 'AAA'
        assert raxml.select_executable_based_on_threads() == 'AAA'
        mock_choice.assert_called_with(['DDD', 'EEE'])

        raxml.threads = 1
        mock_choice.reset_mock()
        mock_choice.return_value = None
        assert raxml.select_executable_based_on_threads() is None
        mock_choice.assert_any_call(['BBB', 'CCC'])
        mock_choice.assert_any_call(['DDD', 'EEE'])

    def test_raxml_convert_raw_ancestral_states_to_fasta(self):
        raxml = treebuilders.RAxML(8)
        raxml.convert_raw_ancestral_states_to_fasta(os.path.join(data_dir, 'raxml_ancestral.state'),
                                                    os.path.join(data_dir, 'raxml_ancestral.fasta.actual'))
        assert filecmp.cmp(os.path.join(data_dir, 'raxml_ancestral.fasta.actual'),
                           os.path.join(data_dir, 'raxml_ancestral.fasta.expected'))
        os.remove(os.path.join(data_dir, 'raxml_ancestral.fasta.actual'))

    def test_raxml_replace_internal_node_label(self):
        raxml = treebuilders.RAxML(8, internal_node_prefix='AAA')
        assert raxml.replace_internal_node_label('Node20') == 'AAANode20'


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import setuptools

setuptools.setup(
    name='gubbins',
    version='2.4.1',
    description='Frontend to the Gubbins BioInformatics tool',
    author='Andrew J. Page',
    author_email='gubbins-help@sanger.ac.uk',
    url='https://github.com/sanger-pathogens/gubbins/',
    packages=setuptools.find_packages(),
    entry_points={
        "console_scripts": [
            "run_gubbins.py = scripts.run_gubbins:main",
        ]
    },
    test_suite='nose.collector',
    tests_require=[
        "nose >= 1.3"
    ],
    long_description="""\
      Gubbins is a tool for BioInformaticians that takes in a multi
      fasta alignment and detects recombination regions.  This package provides
      a simple front end to the Gubbins tool.
      """,
    classifiers=[
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    install_requires=[
        'biopython >= 1.59',
        'dendropy  >= 4.0.2'
    ],
    license='GPL'
)



import sys
import hashlib
from Bio import AlignIO
from Bio.Align import MultipleSeqAlignment
from collections import defaultdict


class PreProcessFasta(object):

    def __init__(self, input_filename, verbose=False, filter_percentage=25):
        self.input_filename = input_filename
        self.verbose = verbose
        self.filter_percentage = filter_percentage

    def hash_sequences(self):
        sequence_hash_to_taxa = defaultdict(list)
        with open(self.input_filename) as input_handle:
            alignments = AlignIO.parse(input_handle, "fasta")
            for alignment in alignments:
                for record in alignment:
                    sequence_hash = hashlib.md5()
                    sequence_hash.update(str(record.seq).encode('utf-8'))
                    hash_of_sequence = sequence_hash.digest()
                    sequence_hash_to_taxa[hash_of_sequence].append(record.id)

                    if self.verbose:
                        print("Sample " + str(record.id) + " has a hash of " + str(hash_of_sequence))
        input_handle.close()
        return sequence_hash_to_taxa

    def calculate_sequences_missing_data_percentage(self):
        sequences_to_missing_data = {}
        with open(self.input_filename) as input_handle:
            alignments = AlignIO.parse(input_handle, "fasta")
            for alignment in alignments:
                for record in alignment:
                    number_of_gaps = 0
                    number_of_gaps += record.seq.count('n')
                    number_of_gaps += record.seq.count('N')
                    number_of_gaps += record.seq.count('-')
                    sequence_length = len(record.seq)

                    if sequence_length == 0:
                        sequences_to_missing_data[record.id] = 100
                        if self.verbose:
                            print("Sample " + str(record.id) + " has no sequence ")
                    else:
                        per_missing_data = number_of_gaps*100/sequence_length
                        sequences_to_missing_data[record.id] = per_missing_data
                        if self.verbose:
                            print("Sample " + str(record.id) + " has missing data percentage of " +
                                  str(per_missing_data))

        input_handle.close()
        return sequences_to_missing_data

    def taxa_missing_too_much_data(self):
        taxa_to_remove = []
        for taxa, percentage_missing in self.calculate_sequences_missing_data_percentage().items():
            if percentage_missing > self.filter_percentage:
                taxa_to_remove.append(taxa)
                print("Excluded sequence " + taxa + " because it had " + str(percentage_missing) +
                      " percentage missing data while a maximum of " + str(self.filter_percentage) + " is allowed")

        return taxa_to_remove

    def taxa_of_duplicate_sequences(self):
        taxa_to_remove = []
        for sequence_hash, taxa in sorted(self.hash_sequences().items()):
            if len(taxa) > 1:
                taxon_to_keep = taxa.pop()
                for taxon in taxa:
                    print("Sequences in " + taxon + " and " + taxon_to_keep + " are identical, removing " + taxon +
                          " from analysis")
                    taxa_to_remove.append(taxon)

        return taxa_to_remove

    def remove_duplicate_sequences_and_sequences_missing_too_much_data(self, output_filename,
                                                                       remove_identical_sequences=None):

        if not remove_identical_sequences:
            taxa_to_remove = self.taxa_missing_too_much_data()
        else:
            taxa_to_remove = self.taxa_of_duplicate_sequences() + self.taxa_missing_too_much_data()

        with open(self.input_filename) as input_handle:
            with open(output_filename, "w+") as output_handle:
                alignments = AlignIO.parse(input_handle, "fasta")
                output_alignments = []

                number_of_included_alignments = 0
                for alignment in alignments:
                    for record in alignment:

                        if record.id not in taxa_to_remove:
                            output_alignments.append(record)
                            number_of_included_alignments += 1

                if number_of_included_alignments <= 1:
                    sys.exit("Not enough sequences are left after removing duplicates.Please check you input data.")

                AlignIO.write(MultipleSeqAlignment(output_alignments), output_handle, "fasta")
                output_handle.close()
            input_handle.close()
        return taxa_to_remove



import os
import re
import sys
from Bio import AlignIO
from collections import Counter
from Bio.Align import MultipleSeqAlignment

class ValidateFastaAlignment(object):
    def __init__(self, input_filename):
      self.input_filename = input_filename

    def is_input_fasta_file_valid(self):
      try:
          if not self.does_each_sequence_have_the_same_length():
              print("Each sequence must be the same length")
              return False
          if not self.are_sequence_names_unique():
              print("All sequence names in the fasta file must be unique")
              return False
          if not self.does_each_sequence_have_a_name_and_genomic_data():
              print("Each sequence must have a name and some genomic data")
              return False
      except:
          return False
    
      return True
    
    def does_each_sequence_have_a_name_and_genomic_data(self):
      with  open(self.input_filename, "r") as input_handle:
        alignments = AlignIO.parse(input_handle, "fasta")
        number_of_sequences = 0
        for alignment in alignments:
            for record in alignment:
                number_of_sequences +=1
                if record.name is None or record.name == "":
                  print("Error with the input FASTA file: One of the sequence names is blank")
                  return False
                if record.seq is None or record.seq == "":
                  print("Error with the input FASTA file: One of the sequences is empty")
                  return False
                if re.search('[^ACGTNacgtn-]', str(record.seq))  != None:
                  print("Error with the input FASTA file: One of the sequences contains odd characters, only ACGTNacgtn- are permitted")
                  return False
        if number_of_sequences <= 3:
          print("Error with input FASTA file: you need more than 3 sequences to build a meaningful tree")
          return False
        input_handle.close()
      return True
    
    def does_each_sequence_have_the_same_length(self):
      try:
        with open(self.input_filename) as input_handle:
    
          alignments = AlignIO.parse(input_handle, "fasta")
          sequence_length = -1
          for alignment in alignments:
              for record in alignment:
                 if sequence_length == -1:
                   sequence_length = len(record.seq)
                 elif sequence_length != len(record.seq):
                   print("Error with the input FASTA file: The sequences dont have the same lengths this isnt an alignment: "+record.name)
                   return False
          input_handle.close()
      except:
        print("Unexpected error:", sys.exc_info()[0])
        print("Error with the input FASTA file: It is in the wrong format so check its an alignment")
        return False
      return True
    
    def are_sequence_names_unique(self):
      with open(self.input_filename) as input_handle:
        alignments = AlignIO.parse(input_handle, "fasta")
        sequence_names = []
        for alignment in alignments:
            for record in alignment:
                sequence_names.append(record.name)
          
        if [k for k,v in list(Counter(sequence_names).items()) if v>1] != []:
          return False
        input_handle.close()
      return True
      


#! /usr/bin/env python3
# encoding: utf-8

"""
Tests the validation of the starting newick tree.
"""

import unittest
from gubbins import common


class TestValidationOfStartingTree(unittest.TestCase):

    def test_are_all_names_unique(self):
        assert not common.is_starting_tree_valid('gubbins/tests/data/duplicate_names_in_tree.tre')

    def test_is_it_a_valid_newick_tree(self):
        assert not common.is_starting_tree_valid('gubbins/tests/data/invalid_newick_tree.tre')
        assert common.is_starting_tree_valid('gubbins/tests/data/valid_newick_tree.tre')

    def test_do_the_names_match_the_fasta_file(self):
        assert common.do_the_names_match_the_fasta_file('gubbins/tests/data/valid_newick_tree.tre',
                                                        'gubbins/tests/data/valid_newick_tree.aln')


if __name__ == "__main__":
    unittest.main()



# encoding: utf-8
# Wellcome Trust Sanger Institute
# Copyright (C) 2013  Wellcome Trust Sanger Institute
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#  
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import os
import sys
import shutil
import pkg_resources
import time
import subprocess
import re
import tempfile

import dendropy
from dendropy.calculate import treecompare

from Bio import AlignIO
from Bio import Phylo
from Bio import SeqIO
from Bio.Align import MultipleSeqAlignment
from Bio.Seq import Seq

from gubbins.PreProcessFasta import PreProcessFasta
from gubbins.ValidateFastaAlignment import ValidateFastaAlignment
from gubbins.treebuilders import FastTree, IQTree, RAxML
from gubbins import utils


def parse_and_run(input_args, program_description=""):
    """Main function of the Gubbins program"""
    start_time = time.time()
    current_directory = os.getcwd()
    printer = utils.VerbosePrinter(True, "\n")

    # Check if the Gubbins C-program is available. If so, print a welcome message. Otherwise exit.
    gubbins_exec = 'gubbins'
    if utils.which(gubbins_exec) is None:
        # Check if the Gubbins C-program is available in its source directory (for tests/Travis)
        gubbins_bundled_exec = os.path.abspath(os.path.join(current_directory, '../src/gubbins'))
        if utils.which(gubbins_bundled_exec) is None:
            sys.exit(gubbins_exec + " is not in your path")
        else:
            gubbins_exec = utils.replace_executable(gubbins_exec, gubbins_bundled_exec)
    program_version = ""
    try:
        program_version = str(pkg_resources.get_distribution(gubbins_exec).version)
    except pkg_resources.RequirementParseError:
        pass
    printer.print(["\n--- Gubbins " + program_version + " ---\n", program_description])

    # Initialize tree builder and ancestral sequence reconstructor; check if all required dependencies are available
    printer.print("\nChecking dependencies...")
    current_tree_name = input_args.starting_tree
    tree_file_names = []
    internal_node_label_prefix = "internal_"
    if input_args.tree_builder == "fasttree" or input_args.tree_builder == "hybrid":
        tree_builder = FastTree(input_args.verbose)
        sequence_reconstructor = RAxML(input_args.threads, input_args.raxml_model, internal_node_label_prefix,
                                       input_args.verbose)
        alignment_suffix = ".snp_sites.aln"
    elif input_args.tree_builder == "raxml":
        tree_builder = RAxML(input_args.threads, input_args.raxml_model, internal_node_label_prefix, input_args.verbose)
        sequence_reconstructor = tree_builder
        alignment_suffix = ".phylip"
    else:
        tree_builder = IQTree(input_args.threads, internal_node_label_prefix, input_args.verbose)
        sequence_reconstructor = tree_builder
        alignment_suffix = ".phylip"
    printer.print("...done. Run time: {:.2f} s".format(time.time() - start_time))

    # Check if the input files exist and have the right format
    printer.print("\nChecking input files...")
    if not os.path.exists(input_args.alignment_filename) \
            or not ValidateFastaAlignment(input_args.alignment_filename).is_input_fasta_file_valid():
        sys.exit("There input alignment file does not exist or has an invalid format")
    if input_args.starting_tree is not None and input_args.starting_tree != "" \
            and (not os.path.exists(input_args.starting_tree) or not is_starting_tree_valid(input_args.starting_tree)):
        sys.exit("The starting tree does not exist or has an invalid format")
    if input_args.starting_tree is not None and input_args.starting_tree != "" \
            and not do_the_names_match_the_fasta_file(input_args.starting_tree, input_args.alignment_filename):
        sys.exit("The names in the starting tree do not match the names in the alignment file")
    if number_of_sequences_in_alignment(input_args.alignment_filename) < 3:
        sys.exit("3 or more sequences are required.")

    # Check - and potentially correct - further input parameters
    check_and_fix_window_size(input_args)

    # Get the base filename
    (base_directory, base_filename) = os.path.split(input_args.alignment_filename)
    (basename, extension) = os.path.splitext(base_filename)
    if input_args.use_time_stamp:
        time_stamp = str(int(time.time()))
        basename = basename + "." + time_stamp
    snp_alignment_filename = base_filename + ".snp_sites.aln"
    gaps_alignment_filename = base_filename + ".gaps.snp_sites.aln"
    gaps_vcf_filename = base_filename + ".gaps.vcf"
    joint_sequences_filename = base_filename + ".seq.joint.aln"

    # Check if intermediate files from a previous run exist
    intermediate_files = [basename + ".iteration_"]
    if not input_args.no_cleanup:
        utils.delete_files(".", intermediate_files, "", input_args.verbose)
    if utils.do_files_exist(".", intermediate_files, "", input_args.verbose):
        sys.exit("Intermediate files from a previous run exist. Please rerun without the --no_cleanup option "
                 "to automatically delete them or with the --use_time_stamp to add a unique prefix.")
    printer.print("...done. Run time: {:.2f} s".format(time.time() - start_time))

    # Filter the input alignment and save as temporary alignment file
    printer.print("\nFiltering input alignment...")
    temp_working_dir = tempfile.mkdtemp(dir=os.getcwd())
    temp_alignment_filename = temp_working_dir + "/" + base_filename

    pre_process_fasta = PreProcessFasta(input_args.alignment_filename, input_args.verbose,
                                        input_args.filter_percentage)
    taxa_removed = pre_process_fasta.remove_duplicate_sequences_and_sequences_missing_too_much_data(
        temp_alignment_filename, input_args.remove_identical_sequences)
    input_args.alignment_filename = temp_alignment_filename

    # If a starting tree has been provided make sure that taxa filtered out in the previous step are removed from it
    if input_args.starting_tree:
        (tree_base_directory, tree_base_filename) = os.path.split(input_args.starting_tree)
        temp_starting_tree = temp_working_dir + '/' + tree_base_filename
        filter_out_removed_taxa_from_tree(input_args.starting_tree, temp_starting_tree, taxa_removed)
        input_args.starting_tree = temp_starting_tree
    printer.print("...done. Run time: {:.2f} s".format(time.time() - start_time))

    # Find all SNP sites with Gubbins
    gubbins_command = " ".join([gubbins_exec, input_args.alignment_filename])
    printer.print(["\nRunning Gubbins to detect SNPs...", gubbins_command])
    try:
        subprocess.check_call(gubbins_command, shell=True)
    except subprocess.SubprocessError:
        sys.exit("Gubbins crashed, please ensure you have enough free memory")
    printer.print("...done. Run time: {:.2f} s".format(time.time() - start_time))
    reconvert_fasta_file(snp_alignment_filename, snp_alignment_filename)
    reconvert_fasta_file(gaps_alignment_filename, base_filename + ".start")

    # Start the main loop
    printer.print("\nEntering the main loop.")
    for i in range(1, input_args.iterations+1):
        printer.print("\n*** Iteration " + str(i) + " ***")

        # 1.1. Construct the tree-building command depending on the iteration and employed options
        if i == 2 and input_args.tree_builder == "hybrid":
            # Switch to RAxML
            tree_builder = sequence_reconstructor
            alignment_suffix = ".phylip"

        if i == 1:
            previous_tree_name = input_args.starting_tree
            alignment_filename = base_filename + alignment_suffix
        else:
            previous_tree_name = current_tree_name
            alignment_filename = previous_tree_name + alignment_suffix

        current_basename = basename + ".iteration_" + str(i)
        current_tree_name = current_basename + ".tre"
        if previous_tree_name:
            tree_building_command = tree_builder.tree_building_command(
                os.path.abspath(alignment_filename), os.path.abspath(previous_tree_name), current_basename)
        else:
            tree_building_command = tree_builder.tree_building_command(
                os.path.abspath(alignment_filename), "", current_basename)
        built_tree = temp_working_dir + "/" + tree_builder.tree_prefix + current_basename + tree_builder.tree_suffix

        # 1.2. Construct the phylogenetic tree
        if input_args.starting_tree is not None and i == 1:
            printer.print("\nCopying the starting tree...")
            shutil.copyfile(input_args.starting_tree, current_tree_name)
        else:
            printer.print(["\nConstructing the phylogenetic tree with " + tree_builder.executable + "...",
                           tree_building_command])
            os.chdir(temp_working_dir)
            try:
                subprocess.check_call(tree_building_command, shell=True)
            except subprocess.SubprocessError:
                sys.exit("Failed while building the tree.")
            os.chdir(current_directory)
            shutil.copyfile(built_tree, current_tree_name)
        printer.print("...done. Run time: {:.2f} s".format(time.time() - start_time))

        # 2. Re-root the tree
        reroot_tree(str(current_tree_name), input_args.outgroup)
        temp_rooted_tree = temp_working_dir + "/" + current_tree_name + ".rooted"
        if input_args.tree_builder == "iqtree":
            shutil.copyfile(current_tree_name, temp_rooted_tree)
        else:
            root_tree(current_tree_name, temp_rooted_tree)

        # 3.1. Construct the command for ancestral state reconstruction depending on the iteration and employed options
        ancestral_sequence_basename = current_basename + ".internal"
        sequence_reconstruction_command = sequence_reconstructor.internal_sequence_reconstruction_command(
            os.path.abspath(base_filename + alignment_suffix), os.path.abspath(temp_rooted_tree),
            ancestral_sequence_basename)
        raw_internal_sequence_filename \
            = temp_working_dir + "/" + sequence_reconstructor.asr_prefix \
            + ancestral_sequence_basename + sequence_reconstructor.asr_suffix
        processed_internal_sequence_filename = temp_working_dir + "/" + ancestral_sequence_basename + ".aln"
        raw_internal_rooted_tree_filename \
            = temp_working_dir + "/" + sequence_reconstructor.asr_tree_prefix \
            + ancestral_sequence_basename + sequence_reconstructor.asr_tree_suffix

        # 3.2. Reconstruct the ancestral sequence
        printer.print(["\nReconstructing ancestral sequences with " + sequence_reconstructor.executable + "...",
                       sequence_reconstruction_command])
        os.chdir(temp_working_dir)
        try:
            subprocess.check_call(sequence_reconstruction_command, shell=True)
        except subprocess.SubprocessError:
            sys.exit("Failed while reconstructing the ancestral sequences.")
        os.chdir(current_directory)

        # 3.3. Join ancestral sequences with given sequences
        current_tree_name_with_internal_nodes = current_tree_name + ".internal"
        sequence_reconstructor.convert_raw_ancestral_states_to_fasta(raw_internal_sequence_filename,
                                                                     processed_internal_sequence_filename)
        concatenate_fasta_files([snp_alignment_filename, processed_internal_sequence_filename],
                                joint_sequences_filename)
        transfer_internal_node_labels_to_tree(raw_internal_rooted_tree_filename, temp_rooted_tree,
                                              current_tree_name_with_internal_nodes, sequence_reconstructor)
        printer.print("...done. Run time: {:.2f} s".format(time.time() - start_time))

        # 4. Reinsert gaps (cp15 note: something is wonky here, the process is at the very least terribly inefficient)
        printer.print("\nReinserting gaps into the alignment...")
        shutil.copyfile(base_filename + ".start", gaps_alignment_filename)
        reinsert_gaps_into_fasta_file(joint_sequences_filename, gaps_vcf_filename, gaps_alignment_filename)
        if not os.path.exists(gaps_alignment_filename) \
                or not ValidateFastaAlignment(gaps_alignment_filename).is_input_fasta_file_valid():
            sys.exit("There is a problem with your FASTA file after running internal sequence reconstruction. "
                     "Please check this intermediate file is valid: " + gaps_alignment_filename)
        printer.print("...done. Run time: {:.2f} s".format(time.time() - start_time))

        # 5. Detect recombination sites with Gubbins (cp15 note: copy file with internal nodes back and forth to
        # ensure all created files have the desired name structure and to avoid fiddling with the Gubbins C program)
        shutil.copyfile(current_tree_name_with_internal_nodes, current_tree_name)
        gubbins_command = create_gubbins_command(
            gubbins_exec, gaps_alignment_filename, gaps_vcf_filename, current_tree_name,
            input_args.alignment_filename, input_args.min_snps, input_args.min_window_size, input_args.max_window_size)
        printer.print(["\nRunning Gubbins to detect recombinations...", gubbins_command])
        try:
            subprocess.check_call(gubbins_command, shell=True)
        except subprocess.SubprocessError:
            sys.exit("Failed while running Gubbins. Please ensure you have enough free memory")
        printer.print("...done. Run time: {:.2f} s".format(time.time() - start_time))
        shutil.copyfile(current_tree_name, current_tree_name_with_internal_nodes)

        # 6. Check for convergence
        printer.print("\nChecking for convergence...")
        remove_internal_node_labels_from_tree(current_tree_name_with_internal_nodes, current_tree_name)
        tree_file_names.append(current_tree_name)
        if i > 1:
            if input_args.converge_method == 'recombination':
                current_recomb_file, previous_recomb_files = get_recombination_files(tree_file_names)
                if have_recombinations_been_seen_before(current_recomb_file, previous_recomb_files):
                    printer.print("Convergence after " + str(i) + " iterations: Recombinations observed before.")
                    break
            else:
                if has_tree_been_seen_before(tree_file_names, input_args.converge_method):
                    printer.print("Convergence after " + str(i) + " iterations: Tree observed before.")
                    break
        printer.print("...done. Run time: {:.2f} s".format(time.time() - start_time))
    else:
        printer.print("Maximum number of iterations (" + str(input_args.iterations) + ") reached.")
    printer.print("\nExiting the main loop.")

    # Create the final output
    printer.print("\nCreating the final output...")
    if input_args.prefix is None:
        input_args.prefix = basename
    output_filenames_to_final_filenames = translation_of_filenames_to_final_filenames(
        current_tree_name, input_args.prefix)
    utils.rename_files(output_filenames_to_final_filenames)

    # Cleanup intermediate files
    if not input_args.no_cleanup:
        shutil.rmtree(temp_working_dir)
        utils.delete_files(".", tree_file_names[:-1], intermediate_files_regex(), input_args.verbose)
        utils.delete_files(".", [base_filename], starting_files_regex(), input_args.verbose)
    printer.print("...finished. Total run time: {:.2f} s".format(time.time() - start_time))


def create_gubbins_command(gubbins_exec, alignment_filename, vcf_filename, current_tree_name,
                           original_alignment_filename, min_snps, min_window_size, max_window_size):
    command = [gubbins_exec, "-r", "-v", vcf_filename, "-a", str(min_window_size),
               "-b", str(max_window_size), "-f", original_alignment_filename, "-t", current_tree_name,
               "-m", str(min_snps), alignment_filename]
    return " ".join(command)


def number_of_sequences_in_alignment(filename):
    return len(get_sequence_names_from_alignment(filename))


def get_sequence_names_from_alignment(filename):
    sequence_names = []
    with open(filename, "r") as handle:
        for record in SeqIO.parse(handle, "fasta"):
            sequence_names.append(record.id)
    return sequence_names


def is_starting_tree_valid(starting_tree):
    try:
        Phylo.read(starting_tree, 'newick')
        dendropy.Tree.get_from_path(starting_tree, 'newick', preserve_underscores=True)
    except Exception:
        print("Error with the input starting tree: Is it a valid Newick file?")
        return False
    return True


def do_the_names_match_the_fasta_file(starting_tree, alignment_filename):
    with open(alignment_filename, "r") as input_handle:
        alignments = AlignIO.parse(input_handle, "fasta")
        sequence_names = {}
        for alignment in alignments:
            for record in alignment:
                sequence_names[record.name] = 1
        input_handle.close()

        tree = dendropy.Tree.get_from_path(starting_tree, 'newick', preserve_underscores=True)

        leaf_nodes = tree.leaf_nodes()
        for i, lf in enumerate(leaf_nodes):
            if not leaf_nodes[i].taxon.label in sequence_names:
                print("Error: A taxon referenced in the starting tree is not found in the input fasta file")
                return False
    return True


def check_and_fix_window_size(input_args):
    if input_args.min_window_size < 3:
        input_args.min_window_size = 3
    if input_args.max_window_size > 1000000:
        input_args.max_window_size = 1000000
    if input_args.min_window_size > input_args.max_window_size:
        input_args.max_window_size, input_args.min_window_size = input_args.min_window_size, input_args.max_window_size


def reconvert_fasta_file(input_filename, output_filename):
    with open(input_filename, "r") as input_handle:
        alignment = AlignIO.read(input_handle, "fasta")
    with open(output_filename, "w+") as output_handle:
        AlignIO.write(alignment, output_handle, "fasta")


def concatenate_fasta_files(input_filenames, output_filename):
    alignments = []
    for input_filename in input_filenames:
        with open(input_filename, "r") as input_handle:
            alignments.append(AlignIO.read(input_handle, "fasta"))

    with open(output_filename, "w+") as output_handle:
        for alignment in alignments:
            AlignIO.write(alignment, output_handle, "fasta")


def starting_files_regex():
    return "\\.(gaps|snp_sites|phylip|vcf|start|seq)"


def intermediate_files_regex():
    return "($|\\.(gff|vcf|snp_sites|branch_snps|phylip|stats|tab|internal))"


def root_tree(input_filename, output_filename):
    # split bi nodes and root tree
    tree = dendropy.Tree.get_from_path(input_filename, 'newick', preserve_underscores=True)
    split_all_non_bi_nodes(tree.seed_node)
    output_tree_string = tree_as_string(tree, suppress_internal=False, suppress_rooting=False)
    with open(output_filename, 'w+') as output_file:
        output_file.write(output_tree_string.replace('\'', ''))


def reroot_tree(tree_name, outgroups):
    if outgroups:
        reroot_tree_with_outgroup(tree_name, outgroups.split(','))
    else:
        reroot_tree_at_midpoint(tree_name)


def reroot_tree_with_outgroup(tree_name, outgroups):
    clade_outgroups = get_monophyletic_outgroup(tree_name, outgroups)
    outgroups = [{'name': taxon_name} for taxon_name in clade_outgroups]

    tree = Phylo.read(tree_name, 'newick')
    tree.root_with_outgroup(*outgroups)
    Phylo.write(tree, tree_name, 'newick')

    tree = dendropy.Tree.get_from_path(tree_name, 'newick', preserve_underscores=True)
    tree.deroot()
    tree.update_bipartitions()
    output_tree_string = tree_as_string(tree, suppress_internal=False)
    with open(tree_name, 'w+') as output_file:
        output_file.write(output_tree_string.replace('\'', ''))


def reroot_tree_at_midpoint(tree_name):
    tree = dendropy.Tree.get_from_path(tree_name, 'newick', preserve_underscores=True)
    split_all_non_bi_nodes(tree.seed_node)
    tree.update_bipartitions()
    tree.reroot_at_midpoint()
    tree.deroot()
    tree.update_bipartitions()
    output_tree_string = tree_as_string(tree, suppress_internal=False)
    with open(tree_name, 'w+') as output_file:
        output_file.write(output_tree_string.replace('\'', ''))


def filter_out_removed_taxa_from_tree(input_filename, output_filename, taxa_removed):
    tree = dendropy.Tree.get_from_path(input_filename, 'newick', preserve_underscores=True)
    tree.prune_taxa_with_labels(taxa_removed, update_bipartitions=True)
    tree.prune_leaves_without_taxa(update_bipartitions=True)
    tree.deroot()
    output_tree_string = tree_as_string(tree)
    with open(output_filename, 'w+') as output_file:
        output_file.write(output_tree_string.replace('\'', ''))


def tree_as_string(tree, suppress_internal=True, suppress_rooting=True):

    return tree.as_string(
        schema='newick',
        suppress_leaf_taxon_labels=False,
        suppress_leaf_node_labels=True,
        suppress_internal_taxon_labels=suppress_internal,
        suppress_internal_node_labels=suppress_internal,
        suppress_rooting=suppress_rooting,
        suppress_edge_lengths=False,
        unquoted_underscores=True,
        preserve_spaces=False,
        store_tree_weights=False,
        suppress_annotations=True,
        annotations_as_nhx=False,
        suppress_item_comments=True,
        node_label_element_separator=' '
    )


def split_all_non_bi_nodes(node):
    if node.is_leaf():
        return None
    elif len(node.child_nodes()) > 2:
        split_child_nodes(node)

    for child_node in node.child_nodes():
        split_all_non_bi_nodes(child_node)

    return None


def split_child_nodes(node):
    all_child_nodes = node.child_nodes()
    # skip over the first node
    first_child = all_child_nodes.pop()
    # create a placeholder node to hang everything else off
    new_child_node = node.new_child(edge_length=0)
    # move the subtree into the placeholder
    new_child_node.set_child_nodes(all_child_nodes)
    # probably not really necessary
    node.set_child_nodes((first_child, new_child_node))


def get_monophyletic_outgroup(tree_name, outgroups):
    if len(outgroups) == 1:
        return outgroups

    tree = dendropy.Tree.get_from_path(tree_name, 'newick', preserve_underscores=True)
    tree.deroot()
    tree.update_bipartitions()

    for leaf_node in tree.mrca(taxon_labels=outgroups).leaf_nodes():
        if leaf_node.taxon.label not in outgroups:
            print("Your outgroups do not form a clade.\n  Using the first taxon " + str(outgroups[0]) +
                  " as the outgroup.\n  Taxon " + str(leaf_node.taxon.label) +
                  " is in the clade but not in your list of outgroups.")
            return [outgroups[0]]

    return outgroups


def transfer_internal_node_labels_to_tree(source_tree_filename, destination_tree_filename, output_tree_filename,
                                          sequence_reconstructor):
    source_tree = dendropy.Tree.get_from_path(source_tree_filename, 'newick', preserve_underscores=True)
    source_internal_node_labels = []
    for source_internal_node in source_tree.internal_nodes():
        if source_internal_node.label:
            source_internal_node_labels.append(source_internal_node.label)
        else:
            source_internal_node_labels.append('')

    destination_tree = dendropy.Tree.get_from_path(destination_tree_filename, 'newick', preserve_underscores=True)
    for index, destination_internal_node in enumerate(destination_tree.internal_nodes()):
        new_label = sequence_reconstructor.replace_internal_node_label(str(source_internal_node_labels[index]))
        destination_internal_node.label = None
        destination_internal_node.taxon = dendropy.Taxon(new_label)

    output_tree_string = tree_as_string(destination_tree, suppress_internal=False, suppress_rooting=False)
    with open(output_tree_filename, 'w+') as output_file:
        output_file.write(output_tree_string.replace('\'', ''))


def remove_internal_node_labels_from_tree(input_filename, output_filename):
    tree = dendropy.Tree.get_from_path(input_filename, 'newick', preserve_underscores=True)
    output_tree_string = tree_as_string(tree)
    with open(output_filename, 'w+') as output_file:
        output_file.write(output_tree_string.replace('\'', ''))


def reinsert_gaps_into_fasta_file(input_fasta_filename, input_vcf_file, output_fasta_filename):
    # find out where the gaps are located
    # PyVCF removed for performance reasons
    with open(input_vcf_file) as vcf_file:

        sample_names = []
        gap_position = []
        gap_alt_base = []

        for vcf_line in vcf_file:
            if re.match('^#CHROM', vcf_line) is not None:
                sample_names = vcf_line.rstrip().split('\t')[9:]
            elif re.match('^\d', vcf_line) is not None:
                # If the alternate is only a gap it wont have a base in this column
                if re.match('^([^\t]+\t){3}([ACGTacgt])\t([^ACGTacgt])\t', vcf_line) is not None:
                    m = re.match('^([^\t]+\t){3}([ACGTacgt])\t([^ACGTacgt])\t', vcf_line)
                    gap_position.append(1)
                    gap_alt_base.append(m.group(2))
                elif re.match('^([^\t]+\t){3}([^ACGTacgt])\t([ACGTacgt])\t', vcf_line) is not None:
                    # sometimes the ref can be a gap only
                    m = re.match('^([^\t]+\t){3}([^ACGTacgt])\t([ACGTacgt])\t', vcf_line)
                    gap_position.append(1)
                    gap_alt_base.append(m.group(3))
                else:
                    gap_position.append(0)
                    gap_alt_base.append('-')

        gapped_alignments = []
        # interleave gap only and snp bases
        with open(input_fasta_filename, "r") as input_handle:
            alignments = AlignIO.parse(input_handle, "fasta")
            for alignment in alignments:
                for record in alignment:
                    inserted_gaps = []
                    if record.id in sample_names:
                        # only apply to internal nodes
                        continue
                    gap_index = 0
                    for input_base in record.seq:
                        while gap_index < len(gap_position) and gap_position[gap_index] == 1:
                            inserted_gaps.append(gap_alt_base[gap_index])
                            gap_index += 1
                        if gap_index < len(gap_position):
                            inserted_gaps.append(input_base)
                            gap_index += 1

                    while gap_index < len(gap_position):
                        inserted_gaps.append(gap_alt_base[gap_index])
                        gap_index += 1

                    record.seq = Seq(''.join(inserted_gaps))
                    gapped_alignments.append(record)

        with open(output_fasta_filename, "a") as output_handle:
            AlignIO.write(MultipleSeqAlignment(gapped_alignments), output_handle, "fasta")
            output_handle.close()
    return


def get_recombination_files(basenames):
    previous_files = []
    for name in basenames:
        previous_files.append(name + ".tab")
    current_file = previous_files.pop()
    return current_file, previous_files


def have_recombinations_been_seen_before(current_file, previous_files):
    if not os.path.exists(current_file):
        return False
    current_file_recombinations = extract_recombinations_from_embl(current_file)

    for previous_file in previous_files:
        if not os.path.exists(previous_file):
            continue
        previous_file_recombinations = extract_recombinations_from_embl(previous_file)
        if current_file_recombinations == previous_file_recombinations:
            return True
    return False


def extract_recombinations_from_embl(filename):
    with open(filename, "r") as fh:
        sequences_to_coords = {}
        start_coord = -1
        end_coord = -1
        for line in fh:
            search_obj = re.search('misc_feature    ([\d]+)..([\d]+)$', line)
            if search_obj is not None:
                start_coord = int(search_obj.group(1))
                end_coord = int(search_obj.group(2))
                continue

            if start_coord >= 0 and end_coord >= 0:
                search_taxa = re.search('taxa\=\"([^"]+)\"', line)
                if search_taxa is not None:
                    taxa_names = search_taxa.group(1).strip().split(' ')
                    for taxa_name in taxa_names:
                        if taxa_name in sequences_to_coords:
                            sequences_to_coords[taxa_name].append([start_coord, end_coord])
                        else:
                            sequences_to_coords[taxa_name] = [[start_coord, end_coord]]

                    start_coord = -1
                    end_coord = -1
                continue
        fh.close()
    return sequences_to_coords


def has_tree_been_seen_before(tree_file_names, converge_method):
    if len(tree_file_names) <= 2:
        return False

    tree_files_which_exist = []
    for tree_file_name in tree_file_names:
        if os.path.exists(tree_file_name):
            tree_files_which_exist.append(tree_file_name)

    for tree_file_name in tree_files_which_exist:
        if tree_file_name is not tree_files_which_exist[-1]:
            if converge_method == 'weighted_robinson_foulds':
                current_rf_distance = robinson_foulds_distance(
                    tree_file_name, tree_files_which_exist[-1])
                if current_rf_distance == 0.0:
                    return True
            else:
                current_rf_distance = symmetric_difference(
                    tree_file_name, tree_files_which_exist[-1])
                if current_rf_distance == 0.0:
                    return True

    return False


def robinson_foulds_distance(input_tree_name, output_tree_name):
    tns = dendropy.TaxonNamespace()
    input_tree = dendropy.Tree.get_from_path(input_tree_name, 'newick', taxon_namespace=tns)
    output_tree = dendropy.Tree.get_from_path(output_tree_name, 'newick', taxon_namespace=tns)
    input_tree.encode_bipartitions()
    output_tree.encode_bipartitions()
    return dendropy.calculate.treecompare.weighted_robinson_foulds_distance(input_tree, output_tree)


def symmetric_difference(input_tree_name, output_tree_name):
    tns = dendropy.TaxonNamespace()
    input_tree = dendropy.Tree.get_from_path(input_tree_name, 'newick', taxon_namespace=tns)
    output_tree = dendropy.Tree.get_from_path(output_tree_name, 'newick', taxon_namespace=tns)
    input_tree.encode_bipartitions()
    output_tree.encode_bipartitions()
    return dendropy.calculate.treecompare.symmetric_difference(input_tree, output_tree)


def translation_of_filenames_to_final_filenames(input_prefix, output_prefix):
    input_names_to_output_names = {
        str(input_prefix) + ".vcf":             str(output_prefix) + ".summary_of_snp_distribution.vcf",
        str(input_prefix) + ".branch_snps.tab": str(output_prefix) + ".branch_base_reconstruction.embl",
        str(input_prefix) + ".tab":             str(output_prefix) + ".recombination_predictions.embl",
        str(input_prefix) + ".gff":             str(output_prefix) + ".recombination_predictions.gff",
        str(input_prefix) + ".stats":           str(output_prefix) + ".per_branch_statistics.csv",
        str(input_prefix) + ".snp_sites.aln":   str(output_prefix) + ".filtered_polymorphic_sites.fasta",
        str(input_prefix) + ".phylip":          str(output_prefix) + ".filtered_polymorphic_sites.phylip",
        str(input_prefix) + ".internal":        str(output_prefix) + ".node_labelled.final_tree.tre",
        str(input_prefix):                      str(output_prefix) + ".final_tree.tre"
    }
    return input_names_to_output_names



# encoding: utf-8
# Wellcome Trust Sanger Institute
# Copyright (C) 2013  Wellcome Trust Sanger Institute
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import os
import shutil
import subprocess
import re


class VerbosePrinter:
    """Class printing messages if verbose argument is set"""

    def __init__(self, verbose: bool, separator: str):
        """Constructor"""
        self.verbose = verbose
        self.sep = separator

    def is_verbose(self):
        return self.verbose

    def separator(self):
        """Get separator"""
        return self.sep

    def make_verbose(self, verbose=True):
        """Set verbosity"""
        self.verbose = verbose

    def set_separator(self, separator: str):
        """Set separator"""
        self.sep = separator

    def print(self, message):
        """Prints a message if set to verbose. If the message is a list, the method prints all elements,
        separated by the set separator"""
        if self.verbose:
            if isinstance(message, list):
                print(*message, sep=self.sep)
            else:
                print(message)


def which(program: str):
    """Checks if a given program exists on the system. Works analogously to the UNIX "which" function"""
    program_and_parameters = program.split(" ")
    if len(program_and_parameters) > 1:
        program = program_and_parameters[0]

    fpath, fname = os.path.split(program)
    if fpath:
        if is_executable(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_executable(exe_file):
                return exe_file
    return None


def is_executable(fpath: str):
    """Checks if a given path corresponds to an executable file"""
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)


def choose_executable(list_of_executables: list):
    """Chooses an executable from a list"""
    for executable in list_of_executables:
        if which(executable) is not None:
            return executable
    return None


def choose_executable_based_on_processor(list_of_executables: list):
    """Chooses an executable from a list and thereby takes into account processor features"""
    flags = []
    cpu_info = False
    if os.path.exists('/proc/cpuinfo'):
        cpu_info = True
        output = subprocess.Popen('grep flags /proc/cpuinfo', stdout=subprocess.PIPE,
                                  shell=True).communicate()[0].decode("utf-8")
        flags = output.split()

    for executable in list_of_executables:
        if cpu_info:
            if 'AVX2' in executable and 'avx2' not in flags:
                continue
            elif 'AVX' in executable and 'avx' not in flags:
                continue
            elif 'SSE3' in executable and 'ssse3' not in flags:
                continue

        if which(executable) is not None:
            return executable

    return None


def replace_executable(command, alternative_executable):
    """Changes the executable in a command"""
    executable_and_params = command.split(" ")
    executable_and_params[0] = alternative_executable
    return " ".join(executable_and_params)


def do_files_exist(directory, basenames, suffix_regex, verbose=False):
    """Checks if files with a given name structure exist"""
    files = os.listdir(directory)
    for file in files:
        full_path = os.path.join(directory, file)
        for basename in basenames:
            regex = "^" + basename + suffix_regex
            if re.match(regex, file) is not None and os.path.exists(full_path):
                if verbose:
                    print("File exists: " + full_path)
                return True
    return False


def delete_files(directory, basenames, suffix_regex, verbose=False):
    """Deletes files with a given name structure"""
    files = os.listdir(directory)
    for file in files:
        full_path = os.path.join(directory, file)
        for basename in basenames:
            regex = "^" + basename + suffix_regex
            if re.match(regex, file) is not None and os.path.exists(full_path):
                if verbose:
                    print("Deleting file: " + full_path)
                os.remove(full_path)


def rename_files(input_to_output_filenames):
    """Renames files"""
    for input_file, output_file in input_to_output_filenames.items():
        if os.path.exists(input_file):
            shutil.move(input_file, output_file)



#! /usr/bin/env python3
# encoding: utf-8

"""
Tests the validation of the input fasta file. It must be a multifasta alignment with sensible data
"""

import unittest
from gubbins.ValidateFastaAlignment import ValidateFastaAlignment


class TestValidateInputFastaFile(unittest.TestCase):

    def test_valid_fasta_file(self):
        validate_fasta = ValidateFastaAlignment('gubbins/tests/data/multiple_recombinations.aln')
        self.assertTrue(validate_fasta.is_input_fasta_file_valid())

    def test_does_each_sequence_have_a_name(self):
        validate_fasta = ValidateFastaAlignment('gubbins/tests/data/sequence_without_a_name.fa')
        self.assertFalse(validate_fasta.does_each_sequence_have_a_name_and_genomic_data())
        self.assertFalse(validate_fasta.is_input_fasta_file_valid())

    def test_does_the_sequence_have_sensible_characters(self):
        validate_fasta = ValidateFastaAlignment('gubbins/tests/data/sequence_with_odd_chars.fa')
        self.assertFalse(validate_fasta.does_each_sequence_have_a_name_and_genomic_data())
        self.assertFalse(validate_fasta.is_input_fasta_file_valid())

    def test_are_all_sequences_the_same_length(self):
        validate_fasta = ValidateFastaAlignment('gubbins/tests/data/valid_alignment.aln')
        self.assertTrue(validate_fasta.does_each_sequence_have_the_same_length())
        self.assertTrue(validate_fasta.is_input_fasta_file_valid())

        validate_fasta = ValidateFastaAlignment('gubbins/tests/data/sequences_of_different_lengths.fa')
        self.assertFalse(validate_fasta.does_each_sequence_have_the_same_length())
        self.assertFalse(validate_fasta.is_input_fasta_file_valid())

    def test_are_all_sequence_names_unique(self):
        validate_fasta = ValidateFastaAlignment('gubbins/tests/data/all_unique_sequence_names.fa')
        self.assertTrue(validate_fasta.are_sequence_names_unique())
        self.assertTrue(validate_fasta.is_input_fasta_file_valid())

        validate_fasta = ValidateFastaAlignment('gubbins/tests/data/non_unique_sequence_names.fa')
        self.assertFalse(validate_fasta.are_sequence_names_unique())
        self.assertFalse(validate_fasta.is_input_fasta_file_valid())

    def test_are_there_enough_sequences_to_build_a_tree(self):
        validate_fasta = ValidateFastaAlignment('gubbins/tests/data/alignment_with_3_sequences.aln')
        self.assertFalse(validate_fasta.is_input_fasta_file_valid())


if __name__ == "__main__":
    unittest.main()
